/* Spa HFP codec API */
/* SPDX-FileCopyrightText: Copyright © 2025 Pauli Virtanen */
/* SPDX-License-Identifier: MIT */

#ifndef SPA_BLUEZ5_HFP_CODEC_CAPS_H_
#define SPA_BLUEZ5_HFP_CODEC_CAPS_H_

#include <spa/param/bluetooth/audio.h>

#define HFP_AUDIO_CODEC_CVSD	0x01
#define HFP_AUDIO_CODEC_MSBC	0x02
#define HFP_AUDIO_CODEC_LC3_SWB	0x03

#define HFP_H2_PACKET_SIZE	60

#endif
