/* PipeWire */
/* SPDX-FileCopyrightText: Copyright © 2022 Wim Taymans */
/* SPDX-License-Identifier: MIT */

#ifndef PULSER_SERVER_CMD_H
#define PULSER_SERVER_CMD_H

#include "internal.h"

int cmd_run(struct impl *impl);

#endif /* PULSER_SERVER_CMD_H */
